import logging
from typing import Any, Dict
from django.http import Http404
from django.core.exceptions import PermissionDenied, ObjectDoesNotExist
from rest_framework import exceptions, status
from rest_framework.response import Response
from rest_framework.views import set_rollback
from vanoma_api_utils.constants import ERROR_CODE
from .responses import generic_error


def _extract_message_from_detail(detail: Any) -> str:
    if isinstance(detail, list):
        errors = [_extract_message_from_detail(d) for d in detail]
        # Return the first error. Subsequent errors will be shown once the user fixes the first one.
        # This is so we can keep the "message" returned in the error as a whole string.
        return errors[0]
    elif isinstance(detail, dict):
        # Return the first error. Subsequent errors will be shown once the user fixes the first one.
        # This is so we can keep the "message" returned in the error as a whole string.
        for key, value in detail.items():
            if str(key).strip().lower() == "non_field_errors":
                return _extract_message_from_detail(value)

            return f"{key}: {_extract_message_from_detail(value)}"

        return errors[0]
    else:
        return str(detail)


def exception_handler(exc: Exception, context: Dict[str, Any]) -> Response:
    """
    Mostly copied from https://github.com/encode/django-rest-framework/blob/master/rest_framework/views.py#L71
    """
    if isinstance(exc, Http404) or isinstance(exc, ObjectDoesNotExist):
        return generic_error(
            status.HTTP_404_NOT_FOUND,
            ERROR_CODE.RESOURCE_NOT_FOUND,
            str(exc),
        )

    if isinstance(exc, PermissionDenied):
        return generic_error(
            status.HTTP_403_FORBIDDEN,
            ERROR_CODE.AUTHORIZATION_ERROR,
            str(exc),
        )

    if isinstance(exc, exceptions.APIException):
        set_rollback()
        return generic_error(
            exc.status_code,
            ERROR_CODE.INVALID_REQUEST,
            _extract_message_from_detail(exc.detail),
        )

    # This will sent the current exception to sentry - https://docs.sentry.io/platforms/python/guides/logging/
    logging.exception(str(exc))

    return generic_error(
        status.HTTP_500_INTERNAL_SERVER_ERROR, ERROR_CODE.INTERNAL_ERROR, str(exc)
    )



from typing import Any, Dict
import os
import jwt
from datetime import datetime, timedelta


class ROLE:
    STAFF = "STAFF"
    DRIVER = "DRIVER"
    AGENT = "AGENT"
    BUYER = "BUYER"
    EMPLOYEE = "EMPLOYEE"
    SERVICE = "SERVICE"


def create_access_token(payload: Dict[str, Any]) -> str:
    return jwt.encode(payload, os.environ["SECRET_KEY"], "HS256")


def create_service_access_token(service_name: str) -> str:
    return create_access_token(
        {
            "sub": service_name,
            "iat": datetime.utcnow(),
            # Service tokens are short-lived since they are generated on-demand
            "exp": datetime.utcnow() + timedelta(seconds=60),
            "roles": [ROLE.SERVICE],
        }
    )



from typing import Dict


from typing import Dict, Any, Union


def before_send(
    event: Dict[str, Any], hint: Dict[str, Any]
) -> Union[Dict[str, Any], None]:
    """Do not sent events generated in testing environmets to avoid hitting our quota limit."""
    if event["environment"] == "testing":
        return None
    return event



from rest_framework.response import Response
from ..constants import ERROR_CODE
from ..misc import create_api_error


def generic_error(status: int, error_code: ERROR_CODE, error_message: str) -> Response:
    return Response(status=status, data=create_api_error(error_code, error_message))



from typing import Any, Dict
from requests import Response
from django.conf import settings
from vanoma_api_utils.http import client
from djangorestframework_camel_case.util import camelize  # type: ignore


class CommunicationApiException(Exception):
    pass


def send_sms(data: Dict[str, Any]) -> Response:
    response = client.post(
        f"{settings.VANOMA_COMMUNICATION_API_URL}/sms", data=camelize(data)
    )

    if not response.ok:
        json = response.json()
        raise CommunicationApiException(json["errorMessage"])

    return response


def send_push(data: Dict[str, Any]) -> Response:
    response = client.post(
        f"{settings.VANOMA_COMMUNICATION_API_URL}/push", data=camelize(data)
    )

    if not response.ok:
        json = response.json()
        raise CommunicationApiException(json["errorMessage"])

    return response


def verify_otp(otp_id: str, otp_code: str, phone_number: str) -> Response:
    response = client.post(
        f"{settings.VANOMA_COMMUNICATION_API_URL}/otp/{otp_id}/verification",
        data={"otpCode": otp_code, "phoneNumber": phone_number},
    )

    if not response.ok:
        json = response.json()
        raise CommunicationApiException(json["errorMessage"])

    return response



from enum import Enum


class ERROR_CODE(Enum):
    INVALID_REQUEST = "INVALID_REQUEST"
    RESOURCE_NOT_FOUND = "RESOURCE_NOT_FOUND"
    AUTHORIZATION_ERROR = "AUTHORIZATION_ERROR"
    INTERNAL_ERROR = "INTERNAL_ERROR"
    PAYMENT_ERROR = "PAYMENT_ERROR"



from random import randint
from faker import Faker
from datetime import datetime


faker = Faker()


def random_phone_number(country_code: str = "250") -> str:
    return f"{country_code}{randint(100000000, 1000000000)}"


def format_datetime(value: datetime) -> str:
    formatted = value.isoformat()

    if formatted.endswith("+00:00"):
        return formatted[:-6] + "Z"

    return formatted



from djangorestframework_camel_case.render import CamelCaseJSONRenderer  # type: ignore


class VanomaMediaTypeRenderer(CamelCaseJSONRenderer):
    """
    We use a custom media type to support versioning. We are overriding the
    media type here so our render can work with our custom media type instead of
    the default application/json.
    """



import uuid
from typing import Any
from django.db import models
from .validators import validate_number
from ..misc import get_shortuuid


class StringField(models.CharField):
    """
    A field used to store a reasonable-sized string. It's a shortcut
    to always having to declare max_length on CharField. Value stored
    in this field should be smaller that what TextField can store but
    also bigger than what ShortStringField (see below) can store.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs["max_length"] = 150
        super().__init__(*args, **kwargs)


class ShortStringField(models.CharField):
    """
    A general purpose field used to store typically single-word type of
    strings. Examples include a primary key value, status and type values, etc.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs["max_length"] = 40
        super().__init__(*args, **kwargs)


class LegacyPrimaryKeyField(ShortStringField):
    """
    Primary key used in auth-api and payment-api services. Because both services
    are used in older delivery services, we have to keep the max length to avoid
    truncating existing rows' primary keys.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs["editable"] = False
        kwargs["primary_key"] = True
        kwargs["default"] = uuid.uuid4
        super().__init__(*args, **kwargs)


class PrimaryKeyField(models.CharField):
    """
    Notice that we are using a shortened version of UUID. This is so that the
    primary key can be used on URL and UUID, although safe, are user-friendly
    in a URL. Among the defaults, max_length is set to 22 as shortuuid generates
    22-character key.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs["editable"] = False
        kwargs["primary_key"] = True
        kwargs["max_length"] = 22
        kwargs["default"] = get_shortuuid
        super().__init__(*args, **kwargs)


class ReferenceKeyField(models.CharField):
    """
    A field that references PrimaryKeyField from another service. In this case
    using models.ForeignKey is not applicable. The convenience of this field
    to set a max length.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs["max_length"] = 22  # Shortuuid are 22 max
        super().__init__(*args, **kwargs)


class PhoneNumberField(models.CharField):
    """
    Field that is used to store phone numbers. It adds validation using Google's
    libphonenumber and sets max length to 15 as suggested by this single -
    https://stackoverflow.com/a/4729239
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs["max_length"] = 15
        kwargs["validators"] = [validate_number]
        super().__init__(*args, **kwargs)



from djangorestframework_camel_case.parser import CamelCaseJSONParser  # type: ignore


class VanomaMediaTypeParser(CamelCaseJSONParser):
    """
    REST framework's header-based versioning looks at the Accept header to determine
    the version. Content-Type is never used to determine version. Therefore, we don't
    have to override media type here as we did for the renderer because clients will
    be include Accept header anywhere. Content-Type (for data submitted for example in
    POST request) will still be application/json.
    """



from typing import Any, Dict, List
from rest_framework import pagination


def create_page(
    count: int, results: List[Any], next: str = None, previous: str = None
) -> Dict[str, Any]:
    return {
        "count": count,
        "next": next,
        "previous": previous,
        "results": results,
    }


class PageNumberPagination(pagination.PageNumberPagination):
    """
    Overriden to allow client to control the page size.
    """

    page_size = 10
    page_size_query_param = "size"



import phonenumbers  # type: ignore


class InvalidPhoneNumber(Exception):
    pass


def is_valid_number(phone_number: str) -> bool:
    if not phone_number:
        return False

    striped_number = str(phone_number).strip()

    if not striped_number.isnumeric():
        return False

    if striped_number.startswith("+"):
        return False

    try:
        # All phone numbers must be in international format with the + sign
        phonenumbers.parse("+{}".format(striped_number), None)
        # TODO: We can truly check for number validity but we are skipping it - https://github.com/daviddrysdale/python-phonenumbers
        return True
    except phonenumbers.NumberParseException:
        return False



from typing import Any, Dict
from django.db.models import Model
from django.core.exceptions import ObjectDoesNotExist
from rest_framework import serializers
from rest_framework.serializers import *  # Import everything so that this file is enough to access DRF stuffs


class BaseModelSerializer(serializers.ModelSerializer):
    """
    Base model serializer which can also be used as a relation field of another serializer. DRF
    separates relation fields from the regular fields of a serializer, but the main difference between
    both from the API'sperspective is that a relation field takes "queryset" parameter. They also differ
    implementation wise in that the to_internal_value method returns an instance of the model in relation
    field while it doesn't for non-relation field. So this class allows model serializer class to be passed
    a queryset parameter to imply that it is being used a relation field.
    """

    def __init__(
        self,
        instance: Model = None,
        data: Any = serializers.empty,
        **kwargs: Any,
    ) -> None:
        self.queryset = kwargs.pop("queryset", None)
        self.primary_key = kwargs.pop("primary_key", None)
        super().__init__(instance=instance, data=data, **kwargs)

    def run_validators(self, value: Any) -> None:
        """
        Serializer inherits from Field and override this method to transform *value* into an iterable
        obviouly because a Serializer contains an iterable of fields. However, if this model is being
        used as a relation field, we expect value to be a non-iterable. In that case, we should call
        the Field.run_validators method instead of Serializer.run_validators
        """
        if self.queryset is None:
            return super().run_validators(value)

        return super(serializers.Serializer, self).run_validators(value)

    def to_internal_value(self, data: Dict[str, Any]) -> Any:
        if self.queryset is None:
            return super().to_internal_value(data)

        if not isinstance(data, dict):
            message = "Invalid data type. Expected a dict but found {}".format(
                type(data)
            )
            raise serializers.ValidationError(message)

        if self.primary_key not in data:
            return super().to_internal_value(data)

        try:
            return self.queryset.get(**{self.primary_key: data[self.primary_key]})
        except ObjectDoesNotExist:
            raise serializers.ValidationError(
                f"Invalid {self.primary_key}. Object does not exist"
            )



from typing import Any, Dict
from djangorestframework_camel_case.util import camelize  # type: ignore


def stringify_filters(filters: Dict[str, Any]) -> str:
    return "&".join(map(lambda item: f"{item[0]}={item[1]}", camelize(filters).items()))



import re
from typing import Any
from django.http.request import HttpRequest
from django.http.response import HttpResponse
from djangorestframework_camel_case.settings import api_settings  # type: ignore
from djangorestframework_camel_case.util import underscoreize  # type: ignore

CAMECAL_CASE_PATTERN = re.compile(r"(?<!^)(?=[A-Z])")


class CamelCaseMiddleWare:
    """
    Middleware to convert camelCase query params to snake_case. Copied from https://github.com/vbabiy/djangorestframework-camel-case/pull/68
    """

    def __init__(self, get_response: Any) -> None:
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:
        request.GET = underscoreize(request.GET, **api_settings.JSON_UNDERSCOREIZE)

        # Convert value of "sort" paremeter to snake case since they refer to fields
        if request.GET.get("sort") is not None:
            request.GET["sort"] = CAMECAL_CASE_PATTERN.sub(
                "_", request.GET["sort"]
            ).lower()

        return self.get_response(request)



from rest_framework import exceptions, status
from django.utils.translation import gettext_lazy as _


class InvalidAPIVersion(exceptions.APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = _("Invalid or missing API version")
    default_code = "invalid_api_version"



from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from vanoma_api_utils.phone_numbers import is_valid_number


def validate_number(phone_number: str) -> None:
    if not is_valid_number(phone_number):
        raise ValidationError(
            _("The value {} is not a valid phone number".format(phone_number))
        )


def validate_numeric(value: str) -> None:
    if not value.isnumeric():
        raise ValidationError(
            _("The value {} is not valid numeric string".format(value))
        )



import json
from typing import Any, Dict, Union
from django.apps import apps
from pathlib import Path


def load_fixture(name: Union[str, None]) -> Union[Dict[str, Any], None]:
    if name is None:
        return None

    for app_config in apps.get_app_configs():
        file_path = f"{app_config.path}/fixtures/{name}"
        if Path(file_path).exists():
            with open(file_path, "r") as rfile:
                return json.load(rfile)

    raise Exception(f"Fixuture {name} not found.")



from typing import Any
from django.http.response import JsonResponse
from rest_framework import status


def create_root_view(api_name: str) -> Any:
    def root_view(*args: Any, **kwargs: Any) -> JsonResponse:
        return JsonResponse(
            {"name": f"Vanoma {api_name.capitalize()} API"}, status=status.HTTP_200_OK
        )

    return root_view



from typing import Any, Dict
from requests import Response
from django.conf import settings
from vanoma_api_utils.http import client
from djangorestframework_camel_case.util import camelize  # type: ignore


class AuthApiException(Exception):
    pass


def create_login(data: Dict[str, Any]) -> Response:
    response = client.post(
        f"{settings.VANOMA_AUTH_API_URL}/login-creation",
        data=camelize(data),
    )

    if not response.ok:
        json = response.json()
        raise AuthApiException(json["errorMessage"])

    return response


def sign_in(data: Dict[str, Any]) -> Response:
    response = client.post(
        f"{settings.VANOMA_AUTH_API_URL}/sign-in",
        data=camelize(data),
    )

    if not response.ok:
        json = response.json()
        raise AuthApiException(json["errorMessage"])

    return response



import os
import shortuuid  # type: ignore
from typing import Dict
from .constants import ERROR_CODE


def resolve_environment() -> str:
    return os.environ.get("ENVIRONMENT", "testing")


def resolve_debug() -> bool:
    return resolve_environment() != "production"


def create_api_error(error_code: ERROR_CODE, error_message: str) -> Dict[str, str]:
    return {"error_code": error_code.value, "error_message": error_message}


def get_shortuuid() -> str:
    return shortuuid.uuid()



from typing import Any, Type
from rest_framework import generics
from rest_framework.serializers import BaseSerializer


class GenericAPIView(generics.GenericAPIView):
    """
    Extends GenericAPIView to add support for resolving serializer class based on the requested API version.
    """

    def get_serializer_class(self, *args: Any, **kwargs: Any) -> Type[BaseSerializer]:
        version_serializer_attr = self._get_version_serializer_attr()

        if version_serializer_attr is None:
            assert (
                self.serializer_class is not None
            ), "serializer_class attribute is required to support version-less requests."

            return self.serializer_class

        assert hasattr(
            self, version_serializer_attr
        ), f"{version_serializer_attr} attribute is missing. It is required to support {self.request.version}."

        return getattr(self, version_serializer_attr)

    def _get_version_serializer_attr(self) -> "str | None":
        if self.request.version is None:
            return None

        return "serializer_class_{}".format(self.request.version.replace(".", "_"))



import os
import dj_database_url  # type: ignore
from typing import Any, Dict
from ..misc import resolve_environment


def resolve_database() -> Dict[str, str]:
    if resolve_environment() == "testing":
        return {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": os.path.join("/tmp", "db.sqlite3"),
        }

    return dj_database_url.config(conn_max_age=600)


def resolve_logging() -> Dict[str, Any]:
    if resolve_environment() == "testing":
        return {}

    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "default": {  # Needed to prettify logs (especially for mini-huey/background task queue) to match gunicon logs format.
                "style": "{",
                "format": "[{asctime}] [{process}] [{levelname}] {message}",
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "default",
            },
        },
        "root": {
            "handlers": ["console"],
            "level": "INFO",
        },
    }


def resolve_file_storage() -> str:
    if resolve_environment() == "testing":
        return "inmemorystorage.InMemoryStorage"

    return "storages.backends.s3boto3.S3Boto3Storage"


BASE_INSTALLED_APPS = [
    "rest_framework",
    "corsheaders",
    "django_filters",
]

BASE_MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.middleware.common.CommonMiddleware",
    "vanoma_api_utils.django.middlewares.CamelCaseMiddleWare",
]

BASE_REST_FRAMEWORK = {
    "ORDERING_PARAM": "sort",
    "UNAUTHENTICATED_USER": None,
    "DEFAULT_FILTER_BACKENDS": (
        "django_filters.rest_framework.DjangoFilterBackend",
        "rest_framework.filters.OrderingFilter",
    ),
    "DEFAULT_RENDERER_CLASSES": (
        "djangorestframework_camel_case.render.CamelCaseJSONRenderer",
    ),
    "DEFAULT_PARSER_CLASSES": (
        "djangorestframework_camel_case.parser.CamelCaseJSONParser",
    ),
    "TEST_REQUEST_DEFAULT_FORMAT": "json",
    "EXCEPTION_HANDLER": "vanoma_api_utils.rest_framework.views.exception_handler",
    "DEFAULT_VERSIONING_CLASS": "rest_framework.versioning.NamespaceVersioning",
    "DEFAULT_PAGINATION_CLASS": "vanoma_api_utils.rest_framework.pagination.PageNumberPagination",
}



from rest_framework import viewsets
from .generics import GenericAPIView


class GenericViewSet(viewsets.ViewSetMixin, GenericAPIView):
    """
    Similar to DRF's viewsets.GenericViewSet except we inherit from our custom
    GenericAPIView to resolve serializer class based on the requested API version.
    """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vanoma_api_utils',
 'vanoma_api_utils.django',
 'vanoma_api_utils.rest_framework']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.2.9,<4.0.0',
 'Faker>=9.9.1,<10.0.0',
 'PyJWT>=2.3.0,<3.0.0',
 'boto3>=1.24.89,<2.0.0',
 'dj-database-url>=0.5.0,<0.6.0',
 'dj-inmemorystorage>=2.1.0,<3.0.0',
 'django-filter>=21.1,<22.0',
 'django-storages>=1.13.1,<2.0.0',
 'djangorestframework-camel-case>=1.2.0,<2.0.0',
 'djangorestframework>=3.12.4,<4.0.0',
 'phonenumbers>=8.12.37,<9.0.0',
 'psycopg2>=2.9.3,<3.0.0',
 'pyhumps>=3.5.3,<4.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'requests>=2.27.1,<3.0.0',
 'sentry-sdk>=1.5.0,<2.0.0',
 'shortuuid>=1.0.9,<2.0.0']

setup_kwargs = {
    'name': 'vanoma-api-utils',
    'version': '0.66.0',
    'description': 'Python utils for vanoma APIs',
    'long_description': None,
    'author': 'Vanoma',
    'author_email': 'contact@vanoma.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import json
import logging
import requests
from typing import Any, Dict, Union
from requests import Response


class Client(object):
    def __init__(self) -> None:
        self.timeout = 20  # 20 seconds to match java client
        self.session = requests.Session()

    def get(self, url: str, headers: Dict[str, Any] = None) -> Response:
        return self._request("GET", url, None, headers)

    def post(
        self, url: str, data: Dict[str, Any], headers: Dict[str, Any] = None
    ) -> Response:
        return self._request("POST", url, data, headers)

    def put(
        self, url: str, data: Dict[str, Any], headers: Dict[str, Any] = None
    ) -> Response:
        return self._request("PUT", url, data, headers)

    def patch(
        self, url: str, data: Dict[str, Any], headers: Dict[str, Any] = None
    ) -> Response:
        return self._request("PATCH", url, data, headers)

    def delete(self, url: str, headers: Dict[str, Any] = None) -> Response:
        return self._request("DELETE", url, None, headers)

    def _request(
        self,
        method: str,
        url: str,
        data: Union[Dict[str, Any], None],
        headers: Union[Dict[str, Any], None],
    ) -> Response:
        response = self.session.request(method, url, json=data, headers=headers)

        if not response.ok:
            try:
                response_body = json.dumps(response.json())
            except Exception:
                response_body = response.text

            logging.warning(
                f"Unexpected response code from: {method} {url} {response.status_code}. RequestHeaders={json.dumps(headers)}. RequestBody={json.dumps(data)}. ResponseHeaders={json.dumps(dict(response.headers))}. ResponseBody={response_body}."
            )

        return response


client = Client()



from rest_framework.test import APIClient, APITestCase


class BaseAPIClient(APIClient):
    ACCEPT_HEADER = "application/json"


class BaseAPITestCase(APITestCase):
    client_class = BaseAPIClient  # type: ignore
